/*************************************************************************************************************************

     Copyright (c) 2015, Analog Devices, Inc.  All rights reserved.

     Redistribution and use in source and binary forms, with or without
     modification, are permitted (subject to the limitations in the
     disclaimer below) provided that the following conditions are met:

     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.

     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the
       distribution.

     * Neither the name of Analog Devices, Inc.  nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

     NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE
     GRANTED BY THIS LICENSE.  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT
     HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
     WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
     LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
     CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
     SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
     BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
     WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
     OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
     IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



**************************************************************************************************************************/


/**************************************************************************************************************************

    File name:    sc57x_init.c
    Purpose:      Initialization code for ADSP-SC58x parts.
    Notes:        Uses the power service to change processor mode and frequency settings.

**************************************************************************************************************************/

#include "sc57x_init.h"

#include <services/pwr/adi_pwr.h>
#include <sys/platform.h>
#include <cdefSC57x_rom.h>
#include <stdint.h>
#include <stdbool.h>
#include "DMCInit.h"
DMC_PARAM DMC_Param_List;
static bool bError = false;

static volatile uint32_t callbackEvents; /* modified in a callback therefore volatile */

static void CheckResult(ADI_PWR_RESULT result, const int32_t lineErrPos);
static void setDefaultPower(uint32_t cgu);
void initcode(ADI_ROM_BOOT_CONFIG* pBootStruct /*unused*/);

/*
 * Note errors if result code is not success.
 */
static inline void CheckResult(ADI_PWR_RESULT result, const int32_t lineErrPos)
{
	static int32_t errCount = 0;
	static int32_t lastErrLinePos;
	if (result != ADI_PWR_SUCCESS) {
		bError = true;
		errCount++;
		lastErrLinePos = lineErrPos;
	}
}

/*
 * Reset processor to default power settings
 */
static void setDefaultPower(uint32_t cgu)
{
	ADI_PWR_RESULT result;

	result = adi_pwr_SetClkControlRegister(cgu, ADI_PWR_CLK_CTL_DF, DF_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkControlRegister(cgu, ADI_PWR_CLK_CTL_MSEL, MSEL_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_SYSSEL, SYSSEL_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_S0SEL, S0SEL_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_S1SEL, S1SEL_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_CSEL, CSEL_DEFAULT);
	CheckResult(result, __LINE__);
	result = adi_pwr_SetClkDivideRegister(cgu, ADI_PWR_CLK_DIV_DSEL, DSEL_DEFAULT);
	CheckResult(result, __LINE__);
} /* setDefaultPower */

/*
 * Power event Callback function
 */
static void PWRCallback(void *pCBParam, uint32_t Event, void *pArg)
{
	uint32_t sclk;
	ADI_PWR_MODE  mode;

	switch ( (ADI_PWR_EVENT)Event )
	{
	case ADI_PWR_EVENT_FREQ_PRE_CHANGE:
		sclk = *(uint32_t*)(pArg);
		callbackEvents++;
		break;
	case ADI_PWR_EVENT_FREQ_POST_CHANGE:
		sclk = *(uint32_t*)(pArg);
		callbackEvents++;
		break;
	case ADI_PWR_EVENT_MODE_PRE_CHANGE:
		mode = (ADI_PWR_MODE)pArg;
		callbackEvents++;
		break;
	case ADI_PWR_EVENT_MODE_POST_CHANGE:
		mode = (ADI_PWR_MODE)pArg;
		callbackEvents++;
		break;
	default:
		break;
	}
} /* PWRCallback */

#if CONFIG_SMPU
static void smpu_config(void)
{
	/* Use SMPU instances to disable accesses to system memory that may not be
	 * populated or needs to be initialized before being accessed. This will
	 * avoid the possibility of an infinite stall in the system due to a
	 * speculative access to a disabled or uninitialized memory. This is also
	 * part of the workaround for silicon anomaly 20000018.
	 */

	/* SMC - SMPU instance 0 */
	*pREG_SMPU0_CTL |= ENUM_SMPU_CTL_RSDIS;    /* Read Speculation Disable */
	*pREG_SMPU0_SECURECTL = /*0xf01*/
					 (ENUM_SMPU_SECURECTL_WSECDIS  /* Disable secure write transactions */
					| ENUM_SMPU_SECURECTL_WNSEN    /* Enable non-secure writes */
					| ENUM_SMPU_SECURECTL_RSECDIS  /* Disable secure read transactions */
					| ENUM_SMPU_SECURECTL_RNSEN    /* Enable non-secure read transactions */
					| ENUM_SMPU_SECURECTL_SBEDIS); /* Disable bus error */


#if CONFIG_DMC0 == DMC_NONE
	/* DMC0 - SMPU instance 9 */
	*pREG_SMPU9_CTL |= ENUM_SMPU_CTL_RSDIS;    /* Read Speculation Disable */
	*pREG_SMPU9_SECURECTL = /*0xf01*/
					 (ENUM_SMPU_SECURECTL_WSECDIS  /* Disable secure write transactions */
					| ENUM_SMPU_SECURECTL_WNSEN    /* Enable non-secure writes */
					| ENUM_SMPU_SECURECTL_RSECDIS  /* Disable secure read transactions */
					| ENUM_SMPU_SECURECTL_RNSEN    /* Enable non-secure read transactions */
					| ENUM_SMPU_SECURECTL_SBEDIS); /* Disable bus error */
#endif

} /* smpu_config */
#endif

void initcode(ADI_ROM_BOOT_CONFIG* pBootStruct /*unused*/)
{
	uint32_t fcclk = 0uL;
	uint32_t fsysclk = 0uL;
	uint32_t fsclk0 = 0uL;
	uint32_t fsclk1 = 0uL;
	uint32_t fdclk = 0uL;

	ADI_PWR_RESULT result;
	ADI_PWR_MODE mode;
	uint32_t expectedEvents = 0uL;

	int32_t i;  /* delay loop counter */
	int32_t fcclk_to_fdclk_ratio;

	bError = false;
	callbackEvents = 0uL;

#if UNUSED_SPI_FLASH
	/* If the SPI Flash memory is not used the following is required to
	 * workaround silicon anomaly 20000018.
	 */
	*pREG_SPI2_TXCTL = (BITM_SPI_TXCTL_TEN | BITM_SPI_TXCTL_TTI);
	*pREG_SPI2_RXCTL = (BITM_SPI_RXCTL_REN | BITM_SPI_RXCTL_RTI);
	*pREG_SPI2_CTL = (BITM_SPI_CTL_EN|BITM_SPI_CTL_MSTR| BITM_SPI_CTL_MMSE);
#endif

#if CONFIG_SMPU
	/* Configure SMPU - also required as a workaround for silicon anomaly 20000018. */
	smpu_config();
#endif

	/* Set the RESETDLL bit (bit 11 of the DMC_PHY_CTL[01] register) before CGU Initialization.*/
#if CONFIG_DMC0
	*pREG_DMC0_PHY_CTL0 |= BITM_DMC_PHY_CTL0_RESETDLL;
#endif

	/* Initialize the CGU to configure the required DCLK frequency. */

	result = adi_pwr_Init(CGU_DEV, CLKIN);
	CheckResult(result, __LINE__);

	result = adi_pwr_InstallCallback(CGU_DEV, PWRCallback);
	CheckResult(result, __LINE__);

#if defined(__ADSPARM__)
	/* Set CGU clock select register to CLKO2/4 (ARM core). */
	result = adi_pwr_SetClkOutSelectRegister(CGU_DEV, ADI_PWR_CLKOUT_CLKO2);
#else
	/* Set CGU clock select register to CLKO0/4 (SHARC core). */
	result = adi_pwr_SetClkOutSelectRegister(CGU_DEV, ADI_PWR_CLKOUT_CLKO0);
#endif
	CheckResult(result, __LINE__);

	/* get the power mode */
	result = adi_pwr_GetPowerMode(CGU_DEV, &mode);
	CheckResult(result, __LINE__);
	if (mode != ADI_PWR_MODE_FULL_ON)
	{
		/* set power mode to full on */
		result = adi_pwr_SetPowerMode(CGU_DEV, ADI_PWR_MODE_FULL_ON);
		CheckResult(result, __LINE__);
		expectedEvents += 2uL; /* pre and post event */

		/* check the power mode is full on */
		result = adi_pwr_GetPowerMode(CGU_DEV, &mode);
		CheckResult(result, __LINE__);
		if (mode != ADI_PWR_MODE_FULL_ON)
		{
			bError = true;
		}
	}

	/* Reset processor to default power settings */
	setDefaultPower(CGU_DEV);

	result = adi_pwr_GetCoreFreq(CGU_DEV, &fcclk);
	CheckResult(result, __LINE__);

	result = adi_pwr_GetSystemFreq(CGU_DEV, &fsysclk, &fsclk0, &fsclk1);
	CheckResult(result, __LINE__);

	result = adi_pwr_GetDDRClkFreq(CGU_DEV, &fdclk);
	CheckResult(result, __LINE__);

	result = adi_pwr_RemoveCallback(CGU_DEV, PWRCallback);
	CheckResult(result, __LINE__);

	/* check the number of callback events */
	if (expectedEvents != callbackEvents) {
		bError = true;
	}

	/* Clear PHY_CTL0 after CGU Initialization. */
#if CONFIG_DMC0
	*pREG_DMC0_PHY_CTL0 = 0uL;
#endif

#if defined(CONFIG_DMC0)
	//Initialize the DMC parameters list
	//Initialize the DMC parameters list
	DMC_Param_List.dmc_no=0;
	DMC_Param_List.ddr_mode=DDR2_MODE;
	DMC_Param_List.phy_init_required=true;
	DMC_Param_List.cclk_dclk_ratio=1;
	DMC_Param_List.anomaly_20000037_applicable=true;
	DMC_Param_List.dmc_ctl_value=DMC_CTL_VALUE;
	DMC_Param_List.dmc_cfg_value=DMC_CFG_VALUE;
	DMC_Param_List.dmc_tr0_value=DMC_TR0_VALUE;
	DMC_Param_List.dmc_tr1_value=DMC_TR1_VALUE;
	DMC_Param_List.dmc_tr2_value=DMC_TR2_VALUE;
	DMC_Param_List.dmc_mr0_value=DMC_MR0_VALUE;
	DMC_Param_List.dmc_mr1_value=DMC_MR1_VALUE;
	DMC_Param_List.dmc_mr2_value=DMC_MR2_VALUE;
	DMC_Param_List.padctl2_value=DMC_PADCTL2_VALUE;
	DMC_Param_List.dmc_cphyctl_value=DMC_CPHYCTL_VALUE;
	DMC_Param_List.dmc_data_calib_add_value=DMC0_DATA_CALIB_ADD;

	//Initialize DMC0 now
	DMCInit(DMC_Param_List);


#endif // __CONFIG_DMC0__


} /* initcode */
